/////////////////////////////////////////////////////////////////////////////
// Name:        TMTiler.cpp
// Purpose:     The entry point for the application, which processes command
//              line arguments, generates the tilesets, and outputs files.
// Author:      Matthew Patitz <mpatitz@yahoo.com>
// Modified by:
// Created:     Nov/2007
// Copyright:   (c) 2008, Matthew John Patitz
// Licence:     GPL
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "TMTiler.h"


TMTileSet::TMTileSet(StateSet& states, Alphabet& alphabet, std::string startState)
{
  m_states      = states;
  m_alphabet    = alphabet;
  m_startState  = startState;
}

void Usage()
{
  printf("\nUsage:  TMTiler.exe -tm <TM definition file> -out <tileset output file> [-seed <seed output file>] [-tilesim] [-nondet]");
  printf("\n        TMTiler.exe -tmpl <template output file>");
  printf("\n");
  printf("\n  -tm      : name of the input file to be used as the Turing machine");
  printf("\n             specification");
  printf("\n  -out     : name of the output file for the generated tileset");
  printf("\n  -seed    : [optional] name of the output file for the generated tile seed");
  printf("\n  -tilesim : [optional] if this option is present, the seed file will be");
  printf("\n             generated in tilesim format, otherwise trend format");
  printf("\n  -tmpl    : name of the output file to be created with a template for a Turing");
  printf("\n             machine specification");
  printf("\n  -nondet  : [optional] if this option is present, the generated tileset will");
  printf("\n             be nondeterministic if the Turing machine halts");
  printf("\n");
  exit(0);
}

int main(int argc, char* argv[])
{
  std::string TMDefFile;
  std::string TileOutFile;
  std::string SeedOutFile;
  bool bTrendSeed = true;
  bool bNonDeterm = false;

  for (int i=1; i<argc; i++) {
    char* foo = argv[i];
    if (strcmp(argv[i], "-tm") == 0) {
      if (argc >= i+1) {
        TMDefFile = argv[++i];
      } else {
        Usage();
      }
    } else if (strcmp(argv[i], "-out") == 0) {
      if (argc >= i+1) {
        TileOutFile = argv[++i];
      } else {
        Usage();
      }
    } else if (strcmp(argv[i], "-seed") == 0) {
      if (argc >= i+1) {
        SeedOutFile = argv[++i];
      } else {
        Usage();
      }
    } else if (strcmp(argv[i], "-tmpl") == 0) {
      if (argc >= i+1) {
        PrintTemplate(argv[++i]);
        exit(0);
      } else {
        Usage();
      }
    } else if (strcmp(argv[i], "-tilesim") == 0) {
      bTrendSeed = false;
    } else if (strcmp(argv[i], "-nondet") == 0) {
      bNonDeterm = true;
    }
  }
  if ((TMDefFile.length() <= 0) || (TileOutFile.length() <= 0)) {
    Usage();
  }

  //////////////////////////////////////////////////////////////////////
  //  Read the Turning machine defintion from the input file

  printf("\nProcessing Turing Machine definition file: '%s'...\n", TMDefFile.c_str());

  std::ifstream fileIn(TMDefFile.c_str());
  if (!fileIn.is_open()) {
    printf("Error:  Unable to open input file '%s'\n", TMDefFile.c_str());
    exit(1);  
  }

  TuringMachine TM;
  std::string   seedStr;

  char lineStr[1001] = {0};
  char *pos = NULL;
  bool bInTransitions = false;
  while (!fileIn.eof()) {
    fileIn.getline(lineStr, 1000);
    std::string tempLine(lineStr);
    tempLine = StripSpaces(tempLine);
    if ((tempLine.length() > 0) && (strncmp(lineStr, "//", 2) != 0)) {    // Skip comment lines and empty lines
      std::vector<std::string> arr;
      if (strncmp(lineStr, STATES_STR, strlen(STATES_STR)) == 0) {
        bInTransitions = false;
        pos = lineStr + strlen(STATES_STR);
        std::string sLine(pos);
        SplitString(sLine, std::string(","), arr);
        for (size_t i=0; i<arr.size(); i++) {
          TM.m_states.push_back(StripSpaces(arr[i]).c_str());
        }
      } else if (strncmp(lineStr, HALTING_STR, strlen(HALTING_STR)) == 0) {
        bInTransitions = false;
        pos = lineStr + strlen(HALTING_STR);
        std::string sLine(pos);
        SplitString(sLine, std::string(","), arr);
        for (size_t i=0; i<arr.size(); i++) {
          TM.m_haltingState = StripSpaces(arr[i]).c_str();
        }
      } else if (strncmp(lineStr, ALPHABET_STR, strlen(ALPHABET_STR)) == 0) {
        bInTransitions = false;
        pos = lineStr + strlen(ALPHABET_STR);
        std::string sLine(pos);
        SplitString(sLine, std::string(","), arr);
        for (size_t i=0; i<arr.size(); i++) {
          std::string s = StripSpaces(arr[i]).c_str();
          if (s == "_") {
            printf("\nError:  The symbol '_' is a reserved character for a blank tape symbol and is included in the tape alphabet which is automatically generated as the input alphabet plus that symbol.  It cannot also be included in the input alphabet.\n\n");
            fileIn.close();
            exit(1);
          }
          TM.m_alphabet.push_back(s);
        }
        TM.m_alphabet.push_back(BLANK);
      } else if (strncmp(lineStr, START_STR, strlen(START_STR)) == 0) {
        bInTransitions = false;
        pos = lineStr + strlen(START_STR);
        std::string sLine(pos);
        TM.m_startState = StripSpaces(sLine);
      } else if (strncmp(lineStr, SEED_STR, strlen(SEED_STR)) == 0) {
        bInTransitions = false;
        pos = lineStr + strlen(SEED_STR);
        std::string sLine(pos);
        seedStr = StripSpaces(sLine);
      } else if (strncmp(lineStr, TRANSITION_STR, strlen(TRANSITION_STR)) == 0) {
        pos = lineStr + strlen(TRANSITION_STR);
        bInTransitions = true;
      } else if (bInTransitions) {
        Transition tran;
        std::string sLine = lineStr;
        size_t mid = sLine.find("->");

        if (mid == -1) {
          printf("\nError:  Invalid transition specification: '%s'\n\n", sLine.c_str());
          fileIn.close();
          exit(1);
        }

        std::string left = sLine.substr(0, mid);
        SplitString(left, std::string(","), arr);
        tran.m_startState = StripSpaces(arr[0]).c_str();
        tran.m_input = StripSpaces(arr[1]);
        arr.clear();

        std::string right = sLine.substr(mid+2, sLine.length());
        SplitString(right, std::string(","), arr);
        tran.m_endState = StripSpaces(arr[0]).c_str();
        tran.m_output = StripSpaces(arr[1]).c_str();
        std::string sDir = StripSpaces(arr[2]);
        if (sDir.compare("r")==0 || sDir.compare("R")==0) {
          tran.m_direction = RIGHT;
        } else if (sDir.compare("l")==0 || sDir.compare("L")==0) {
          tran.m_direction = LEFT;
        } else {
          tran.m_direction = INVALID;
        }

        if (tran.IsValid()) {
          TM.m_transitions.push_back(tran);
        } else {
          printf("Error:  Invalid transition: %s\n", sLine);
        }
      }
    }
  }
  fileIn.close();

  if (!TM.IsValid()) {
    printf("Error:  Invalid Turing machine definition!\n");
    exit(1);
  }


  //////////////////////////////////////////////////////////////////////
  //  Create the equivalent tile set from the Turing machine's definition

  TMTileSet ts(TM.m_states, TM.m_alphabet, TM.m_startState);

  //  Create tiles for:
  //  1.  copy
  //  2.  start state in seed row
  //  3.  Tau-stable seed

  for (size_t k=0; k<TM.m_alphabet.size(); k++) {
    Tile a;
    a.m_display                 = TM.m_alphabet[k];
    a.m_sides[NORTH].m_strength = 1;
    a.m_sides[EAST].m_strength  = 1;
    a.m_sides[SOUTH].m_strength = 1;
    a.m_sides[WEST].m_strength  = 1;
    a.m_sides[NORTH].m_label    = TM.m_alphabet[k];
    a.m_sides[EAST].m_label     = COPYLEFT;
    a.m_sides[SOUTH].m_label    = TM.m_alphabet[k];
    a.m_sides[WEST].m_label     = COPYLEFT;
    ts.push_back(a);

    Tile a2;
    a2.m_display                 = TM.m_alphabet[k];
    a2.m_sides[NORTH].m_strength = 1;
    a2.m_sides[EAST].m_strength  = 1;
    a2.m_sides[SOUTH].m_strength = 1;
    a2.m_sides[WEST].m_strength  = 1;
    a2.m_sides[NORTH].m_label    = TM.m_alphabet[k];
    a2.m_sides[EAST].m_label     = COPYRIGHT;
    a2.m_sides[SOUTH].m_label    = TM.m_alphabet[k];
    a2.m_sides[WEST].m_label     = COPYRIGHT;
    ts.push_back(a2);

    std::string bName = std::string("[") + TM.m_startState + "]," + TM.m_alphabet[k].c_str() + std::string("SD");
    Tile b(bName.c_str());
    b.m_display                 = "[" + TM.m_startState + "]," + TM.m_alphabet[k];
    b.m_sides[NORTH].m_strength = 2;
    b.m_sides[EAST].m_strength  = 2;
    b.m_sides[SOUTH].m_strength = 0;
    b.m_sides[WEST].m_strength  = 0;
    b.m_sides[NORTH].m_label    = "[" + TM.m_startState + "]," + TM.m_alphabet[k];
    b.m_sides[EAST].m_label     = std::string("SD") + COPYRIGHT;
    b.m_sides[SOUTH].m_label    = NODISPLAY;
    b.m_sides[WEST].m_label     = NODISPLAY;
    ts.push_back(b);

    std::string cName = TM.m_alphabet[k] + "SD";
    Tile c(cName.c_str());
    c.m_display                 = TM.m_alphabet[k];
    c.m_sides[NORTH].m_strength = 1;
    c.m_sides[EAST].m_strength  = 2;
    c.m_sides[SOUTH].m_strength = 0;
    c.m_sides[WEST].m_strength  = 2;
    c.m_sides[NORTH].m_label    = TM.m_alphabet[k];
    c.m_sides[EAST].m_label     = std::string("SD") + COPYRIGHT;
    c.m_sides[SOUTH].m_label    = NODISPLAY;
    c.m_sides[WEST].m_label     = std::string("SD") + COPYRIGHT;
    ts.push_back(c);
  }

  //  Right side ('space') tiles
  Tile right1("RIGHT0");
  right1.m_display                  = BLANK;
  right1.m_sides[NORTH].m_strength  = 1;
  right1.m_sides[EAST].m_strength   = 0;
  right1.m_sides[SOUTH].m_strength  = 0;
  right1.m_sides[WEST].m_strength   = 1;
  right1.m_sides[NORTH].m_label     = std::string(BLANK) + "*";
  right1.m_sides[EAST].m_label      = NODISPLAY;
  right1.m_sides[SOUTH].m_label     = NODISPLAY;
  right1.m_sides[WEST].m_label      = COPYRIGHT;
  ts.push_back(right1);

  Tile rightsd("RIGHTSD");
  rightsd.m_display                  = BLANK;
  rightsd.m_sides[NORTH].m_strength  = 1;
  rightsd.m_sides[EAST].m_strength   = 0;
  rightsd.m_sides[SOUTH].m_strength  = 0;
  rightsd.m_sides[WEST].m_strength   = 2;
  rightsd.m_sides[NORTH].m_label     = std::string("SD") + BLANK + "*";
  rightsd.m_sides[EAST].m_label      = NODISPLAY;
  rightsd.m_sides[SOUTH].m_label     = NODISPLAY;
  rightsd.m_sides[WEST].m_label      = COPYRIGHT;
  ts.push_back(rightsd);

  Tile right2("RIGHT1");
  right2.m_display                  = BLANK;
  right2.m_sides[NORTH].m_strength  = 1;
  right2.m_sides[EAST].m_strength   = 2;
  right2.m_sides[SOUTH].m_strength  = 1;
  right2.m_sides[WEST].m_strength   = 1;
  right2.m_sides[NORTH].m_label     = BLANK;
  right2.m_sides[EAST].m_label      = std::string(BLANK) + "*";
  right2.m_sides[SOUTH].m_label     = std::string(BLANK) + "*";
  right2.m_sides[WEST].m_label      = COPYRIGHT;
  ts.push_back(right2);

  Tile right3("RIGHT2");
  right3.m_display                  = BLANK;
  right3.m_sides[NORTH].m_strength  = 1;
  right3.m_sides[EAST].m_strength   = 0;
  right3.m_sides[SOUTH].m_strength  = 0;
  right3.m_sides[WEST].m_strength   = 2;
  right3.m_sides[NORTH].m_label     = std::string(BLANK) + "*";
  right3.m_sides[EAST].m_label      = NODISPLAY;
  right3.m_sides[SOUTH].m_label     = NODISPLAY;
  right3.m_sides[WEST].m_label      = std::string(BLANK) + "*";
  ts.push_back(right3);

  for (size_t t=0; t<TM.m_transitions.size(); t++) {
    std::string endState = TM.m_transitions[t].m_endState;

    for (size_t t2=0; t2<TM.m_transitions.size(); t2++) {
      std::string startState = TM.m_transitions[t2].m_startState;

      if (startState == endState) {

        std::string northLabel = "[" + endState + "]," + TM.m_transitions[t2].m_input;

        for (size_t t3=0; t3<TM.m_transitions.size(); t3++) {
          if ((TM.m_transitions[t3].m_startState == endState) &&
              (TM.m_transitions[t3].m_input == TM.m_transitions[t2].m_input))
          {
            if (TM.m_transitions[t3].m_endState == TM.m_haltingState) {
              northLabel = "HALT";
              break;
            }
          }
        }

        if (TM.m_transitions[t].m_direction == LEFT) {
          Tile m;
          m.m_display                 = "[" + endState + "]," + TM.m_transitions[t2].m_input;
          m.m_sides[NORTH].m_strength = 2;
          m.m_sides[EAST].m_strength  = 1;
          m.m_sides[SOUTH].m_strength = 1;
          m.m_sides[WEST].m_strength  = 1;
          m.m_sides[NORTH].m_label    = northLabel;
          m.m_sides[EAST].m_label     = "[" + TM.m_transitions[t].m_startState + "]," + TM.m_transitions[t].m_input;
          m.m_sides[SOUTH].m_label    = TM.m_transitions[t2].m_input;
          m.m_sides[WEST].m_label     = COPYLEFT;
          ts.push_back(m); 
        } else {
          Tile m;
          m.m_display                 = "[" + endState + "]," + TM.m_transitions[t2].m_input;
          m.m_sides[NORTH].m_strength = 2;
          m.m_sides[EAST].m_strength  = 1;
          m.m_sides[SOUTH].m_strength = 1;
          m.m_sides[WEST].m_strength  = 1;
          m.m_sides[NORTH].m_label    = northLabel;
          m.m_sides[EAST].m_label     = COPYRIGHT;
          m.m_sides[SOUTH].m_label    = TM.m_transitions[t2].m_input;
          m.m_sides[WEST].m_label     = "[" + TM.m_transitions[t].m_startState + "]," + TM.m_transitions[t].m_input;
          ts.push_back(m);
        }
      }
    }
    
    if (TM.m_transitions[t].m_direction == LEFT) {
      Tile top;
      top.m_display                 = TM.m_transitions[t].m_output;
      top.m_sides[NORTH].m_strength = 1;
      top.m_sides[EAST].m_strength  = 1;
      top.m_sides[SOUTH].m_strength = 2;
      top.m_sides[WEST].m_strength  = 1;
      top.m_sides[NORTH].m_label    = TM.m_transitions[t].m_output;
      top.m_sides[EAST].m_label     = COPYRIGHT;
      top.m_sides[SOUTH].m_label    = "[" + TM.m_transitions[t].m_startState + "]," + TM.m_transitions[t].m_input;
      top.m_sides[WEST].m_label     = "[" + TM.m_transitions[t].m_startState + "]," + TM.m_transitions[t].m_input;
      ts.push_back(top);
    } else {
      Tile top;
      top.m_display                 = TM.m_transitions[t].m_output;
      top.m_sides[NORTH].m_strength = 1;
      top.m_sides[EAST].m_strength  = 1;
      top.m_sides[SOUTH].m_strength = 2;
      top.m_sides[WEST].m_strength  = 1;
      top.m_sides[NORTH].m_label    = TM.m_transitions[t].m_output;
      top.m_sides[EAST].m_label     = "[" + TM.m_transitions[t].m_startState + "]," + TM.m_transitions[t].m_input;
      top.m_sides[SOUTH].m_label    = "[" + TM.m_transitions[t].m_startState + "]," + TM.m_transitions[t].m_input;
      top.m_sides[WEST].m_label     = COPYLEFT;
      ts.push_back(top);
    }
  }

  if (bNonDeterm) {
    printf("\nGenerating tiletypes for a nondeterministic Turing machine assembly\n");
  }

  //  Halting state tile
  Tile halt("HALT");
  halt.m_display                 = "HALT";
  halt.m_sides[NORTH].m_strength = 0;
  halt.m_sides[EAST].m_strength  = 0;
  halt.m_sides[SOUTH].m_strength = 2;
  halt.m_sides[WEST].m_strength  = 0;
  halt.m_sides[NORTH].m_label    = NODISPLAY;
  halt.m_sides[EAST].m_label     = NODISPLAY;
  halt.m_sides[SOUTH].m_label    = "HALT";
  halt.m_sides[WEST].m_label     = NODISPLAY;
  halt.m_color                   = "RGB(255,0,0)";
  ts.push_back(halt);

  if (bNonDeterm) {
    Tile halt2("HALT2");
    halt2.m_display                 = "HALT2";
    halt2.m_sides[NORTH].m_strength = 0;
    halt2.m_sides[EAST].m_strength  = 0;
    halt2.m_sides[SOUTH].m_strength = 2;
    halt2.m_sides[WEST].m_strength  = 0;
    halt2.m_sides[NORTH].m_label    = NODISPLAY;
    halt2.m_sides[EAST].m_label     = NODISPLAY;
    halt2.m_sides[SOUTH].m_label    = "HALT";
    halt2.m_sides[WEST].m_label     = NODISPLAY;
    halt2.m_color                   = "RGB(0,0,255)";
    ts.push_back(halt2);
  }

  //////////////////////////////////////////////////////////////////////
  //  Output the tileset

  std::ofstream out(TileOutFile.c_str());
  if (!out.is_open()) {
    printf("Error:  Failed to open the output file '%s'\n", TileOutFile.c_str());
    exit(1);
  }

  printf("\nOutputting tileset to file: '%s'...\n\n", TileOutFile.c_str());
  for (size_t j=0; j<ts.size(); j++) {
    ts[j].Print(&out);
    out << "\n";
  }
  out.close();

  if (seedStr.length() == 0) {
    printf("TMTiler has completed successfully.\n");
	  return 0;
  }


  //////////////////////////////////////////////////////////////////////
  //  Create the set of seed tiles

  TileSet seedSet;
  std::vector<std::string> arr;
  SplitString(seedStr, std::string(","), arr);
  bool bFirst = true;
  for (size_t st=0; st<arr.size(); st++) {
    std::string sc(StripSpaces(arr[st]));
    bool bOkay = false;
    for (size_t ac=0; ac<TM.m_alphabet.size(); ac++) {
      if (sc == TM.m_alphabet[ac]) {
        bOkay = true;
        break;
      }
    }
    if (!bOkay) {
      printf("Error:  Invalid alphabet character '%s' specified for seed entry!\n", sc.c_str());
      exit(1);
    }
    bool bFoundIt = false;
    if (bFirst) {
      for (size_t ttnum=0; ttnum<ts.size(); ttnum++) {
        std::string tileName = "[" + TM.m_startState + "]," + sc + "SD";
        if (ts[ttnum].GetName() == tileName) {
          seedSet.push_back(ts[ttnum]);
          bFoundIt = true;
          break;
        }
      }
      bFirst = false;
    } else {
      for (size_t ttnum=0; ttnum<ts.size(); ttnum++) {
        std::string tileName = sc + "SD";
        if (ts[ttnum].GetName() == tileName) {
          seedSet.push_back(ts[ttnum]);
          bFoundIt = true;
          break;
        }
      }
    }
    if (!bFoundIt) {
      printf("Error:  Couldn't find correct tile type for seed tile with symbol '%s'\n", sc.c_str());
    }
  }


  //////////////////////////////////////////////////////////////////////
  //  Output the seed definition

  std::ofstream seedOut(SeedOutFile.c_str());
  if (!seedOut.is_open()) {
    printf("Error:  Failed to open the seed output file '%s'\n", SeedOutFile);
    exit(1);
  }

  std::string tdpFile = TileOutFile.substr((TileOutFile.find_last_of('\\')+1), TileOutFile.length());
  seedOut << tdpFile;
  seedOut << "\n";
  for (size_t snum=0; snum<seedSet.size(); snum++) {
    seedOut << seedSet[snum].GetName();
    seedOut << " ";
    if (bTrendSeed) {
      seedOut << (snum+1);
      seedOut << " ";
      seedOut << "1\n";
    } else {
      seedOut << "1 ";
      seedOut << (snum+1);
      seedOut << "\n";
    }
  }
  if (bTrendSeed) {
    seedOut << "RIGHT0 ";
    seedOut << (seedSet.size()+1);
    seedOut << " 1\n";
  } else {
    seedOut << "RIGHT0 1 ";
    seedOut << (seedSet.size()+1);
    seedOut << "\n";
  }

  printf("Outputting seed definition to file: '%s'...\n\n", SeedOutFile.c_str());

  printf("TMTiler has completed successfully.\n");
	return 0;
}

void PrintTemplate(const char* tmplFileName)
{
  std::ofstream out(tmplFileName);
  if (!out.is_open()) {
    printf("Error:  Failed to open the template output file '%s'\n", tmplFileName);
    exit(1);
  }

  printf("\nOutputting template to file: '%s'...\n\n", tmplFileName);

  out << "/////////////////////////////////////////////////////////////////////////////\n";
  out << "//  Turing Machine Specification\n";
  out << "//    Assumptions:\n";
  out << "//      - the tape is one-way infinite to the right\n";
  out << "//      - the TM will begin in the start state with the tape head\n";
  out << "//        on the leftmost cell of the tape\n";
  out << "//      - the tape alphabet will be automatically generated as\n";
  out << "//        exactly the input plus the '_' symbol to designate blanks\n";
  out << "//    Reserved characters:\n";
  out << "//      - These symbols are used by TMTiler in generated names and\n";
  out << "//        therefore can't appear in the input alphabet or state names:\n";
  out << "//          ";
  int i = 0;
  while (ReservedChars[i]) {
    if (i > 0) {
      out << ", ";
    }
    out << std::string("'") + ReservedChars[i] + std::string("'");
    i++;
  }
  out << "\n";
  out << "//    Naming convention:\n";
  out << "//      - TMTiler will automatically generate tile names for most\n";
  out << "//        tiles as a concatention of four edge labels as follows:\n";
  out << "//          <north label>|<east lable>|<south label>|<west label>\n";
  out << "/////////////////////////////////////////////////////////////////////////////\n";
  out << "\n";
  out << "//  States:  a comma separated list of state names (e.g. \"q0, q1, q2\")\n";
  out << "States: \n";
  out << "\n";
  out << "//  Alphabet:  a comma separated list of alphabet symbols (e.g. \"0, 1, 2\")\n";
  out << "Alphabet: \n";
  out << "\n";
  out << "//  Start state: the name of the state to be treated as the start state\n";
  out << "Start state: \n";
  out << "\n";
  out << "//  Halting state:  the name of the state to be treated as the halting state (e.g. \"q5\")\n";
  out << "Halting state: \n";
  out << "\n";
  out << "//  Transitions:  a list of transitions, one per line, in the following form\n";
  out << "//    <state>, <input> -> <state>, <output>, <direction: L|R>\n";
  out << "//    Example:  q0, 0 -> q1, 1, R\n";
  out << "Transitions:\n";
  out << "  \n";
  out << "\n";
  out << "/////////////////////////////////////////////////////////////////////////////\n";
  out << "//  Seed: [Optional] a comma separated list of alphabet values representing\n";
  out << "//        the tape values starting from the leftmost to the rightmost,\n";
  out << "//        (e.g. \"0, 0, 1, 0, 1, 1, 0\")\n";
  out << "//        Assume a blank ('_') will be automatically appended to the right.\n";
  out << "//        Assume the simulation starts in the start state with the tape head\n";
  out << "//        on the leftmost tape cell.\n";
  out << "Seed: \n";

  out.close();
  printf("Completed successfully.\n");
}
